import datetime


def check_url_type(url):
    if type(url) != str:
        raise TypeError("Have you checked your url connection string is correct?")


def check_log_type(log):
    if type(log) is not dict:
        raise TypeError(
            f"Have you checked your log is a dictionary?\nlog must be a dictionary with keys: datetime_utc, input, output, total_tokens"
        )
    if set(log.keys()) != {"datetime_utc", "input", "output", "total_tokens"}:
        raise KeyError(
            "log must be a dictionary with keys: datetime_utc, input, output, total_tokens"
        )
    if type(log["datetime_utc"]) is not datetime.datetime:
        raise TypeError("log.datetime_utc must be a datetime object")
    if type(log["input"]) is not str:
        raise TypeError("log.input must be a string")
    if type(log["output"]) is not str:
        raise TypeError("log.output must be a string")
    if type(log["total_tokens"]) is not int:
        raise TypeError("log.total_tokens must be an integer")



from datetime import datetime
import json
import logging
import sys
import openai
from send_to_db import send_to_db


def start_listener(url: str):
    openai.log = "debug"

    log_handler = logging.StreamHandler()
    log_handler.addFilter(OpenAIFilter(url))

    logger = logging.getLogger("omnilog")
    logger.handlers = [log_handler]

    sys.stderr = StreamToLogger(logger, logging.ERROR)


def handle_openai_response(data: str, url: str):
    output = data.split("body='")[1].split("' headers=")[0]
    output = bytes(output, "utf-8").decode("unicode_escape")
    output = json.loads(output)

    log = {
        "input": "This is a new dummy input!",
        "datetime_utc": datetime.utcfromtimestamp(output["created"]),
        "output": output["choices"][0]["text"],
        "total_tokens": output["usage"]["total_tokens"],
    }
    send_to_db(url, log)


class OpenAIFilter(logging.Filter):
    url: str

    def __init__(self, url):
        self.url = url

    def filter(self, record) -> bool:
        msg = record.getMessage()

        if "OpenAI" in msg or "openai" in msg:
            if "message='Request to OpenAI API'" in msg:
                return False

            if "message='Post details'" in msg:
                return False

            if "message='OpenAI API response'" in msg:
                return False

            if "message='API response body'" in msg:
                try:
                    handle_openai_response(msg, self.url)
                except json.decoder.JSONDecodeError:
                    record.msg = "JSONDecodeError for OpenAI response: " + msg
                    return True
                return False
        return True


class StreamToLogger(object):
    """
    Fake file-like stream object that redirects writes to a logger instance.
    """

    def __init__(self, logger, log_level=logging.INFO):
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ""

    def write(self, buf):
        temp_linebuf = self.linebuf + buf
        self.linebuf = ""
        for line in temp_linebuf.splitlines(True):
            # From the io.TextIOWrapper docs:
            #   On output, if newline is None, any '\n' characters written
            #   are translated to the system default line separator.
            # By default sys.stdout.write() expects '\n' newlines and then
            # translates them so this is still cross platform.
            if line[-1] == "\n":
                self.logger.log(self.log_level, line.rstrip())
            else:
                self.linebuf += line

    def flush(self):
        if self.linebuf != "":
            self.logger.log(self.log_level, self.linebuf.rstrip())
        self.linebuf = ""



import psycopg2
from input_validation import check_log_type, check_url_type


def send_to_db(url, log):
    check_url_type(url)
    check_log_type(log)

    connection, cursor = connect_to_db(url)

    cursor.execute(
        """
    INSERT INTO llm_logs (datetime_utc, input_string, output_string, total_tokens)
    VALUES (%s, %s, %s, %s);
    """,
        (log["datetime_utc"], log["input"], log["output"], log["total_tokens"]),
    )

    connection.commit()
    cursor.close()
    connection.close()


def connect_to_db(url):
    try:
        conn = psycopg2.connect(url)
    except psycopg2.OperationalError:
        print(f"Could not connect to database: {url}\n")
        raise

    cur = conn.cursor()
    cur.execute("SELECT NOW();")
    time = cur.fetchone()[0]

    cur.execute("SELECT version();")
    version = cur.fetchone()[0]

    if not time or not version:
        raise ConnectionError(f"Could not connect to database: {url}")

    print("Current time:", time)
    print("PostgreSQL version:", version)
    return conn, cur
